package de.ozgcloud.apilib.vorgang.grpc;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.List;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.apilib.vorgang.OzgCloudVorgangQuery;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangQueryTestFactory;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangStub;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangStubTestFactory;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangTestFactory;
import de.ozgcloud.apilib.vorgang.Page;
import de.ozgcloud.vorgang.vorgang.GrpcFindVorgangRequest;
import de.ozgcloud.vorgang.vorgang.GrpcFindVorgangResponse;
import de.ozgcloud.vorgang.vorgang.GrpcFindVorgangWithEingangRequest;
import de.ozgcloud.vorgang.vorgang.GrpcVorgangHeader;
import de.ozgcloud.vorgang.vorgang.VorgangServiceGrpc.VorgangServiceBlockingStub;

class GrpcOzgCloudVorgangServiceTest {

	@InjectMocks
	@Spy
	private GrpcOzgCloudVorgangService service;

	@Mock
	private VorgangServiceBlockingStub stub;
	@Mock
	private OzgCloudVorgangMapper mapper;
	@Mock
	private OzgCloudVorgangStubMapper stubMapper;
	@Mock
	private OzgCloudVorgangQueryMapper queryMapper;

	@Nested
	class TestGetById {
		@Captor
		private ArgumentCaptor<GrpcFindVorgangWithEingangRequest> requestCaptor;

		@BeforeEach
		void init() {
			when(stub.withInterceptors(any())).thenReturn(stub);
			when(stub.findVorgangWithEingang(any())).thenReturn(GrpcFindVorgangWithEingangResponseTestFactory.create());
		}

		@Test
		void shouldCallStub() {
			service.getById(OzgCloudVorgangTestFactory.ID);

			verify(stub).findVorgangWithEingang(notNull());
		}

		@Test
		void shouldHaveVorgangId() {
			service.getById(OzgCloudVorgangTestFactory.ID);

			verify(stub).findVorgangWithEingang(requestCaptor.capture());
			assertThat(requestCaptor.getValue().getId()).isEqualTo(OzgCloudVorgangTestFactory.ID.toString());
		}

		@Test
		void shouldMapResponse() {
			service.getById(OzgCloudVorgangTestFactory.ID);

			verify(mapper).fromGrpc(any());
		}

		@Test
		void shouldReturnMappedVorgang() {
			var vorgang = OzgCloudVorgangTestFactory.create();
			when(mapper.fromGrpc(any())).thenReturn(vorgang);

			var result = service.getById(OzgCloudVorgangTestFactory.ID);

			assertThat(result).isSameAs(vorgang);
		}
	}

	@Nested
	class TestSearch {

		private static final String SEARCH_STR = "search";

		@Captor
		private ArgumentCaptor<GrpcFindVorgangRequest> requestCaptor;

		@Nested
		class BuildSearchRequest {
			private static final Page PAGE = Page.builder().limit(40).offset(21).build();

			@Test
			void shouldSetSearchString() {
				var request = service.buildSearchRequest(SEARCH_STR, PAGE);

				assertThat(request.getSearchBy()).isEqualTo(SEARCH_STR);
			}

			@Test
			void shouldSetLimit() {
				var request = service.buildSearchRequest(SEARCH_STR, PAGE);

				assertThat(request.getLimit()).isEqualTo(40);
			}

			@Test
			void shouldSetOffset() {
				var request = service.buildSearchRequest(SEARCH_STR, PAGE);

				assertThat(request.getOffset()).isEqualTo(21);
			}
		}

		@Nested
		class doingRequest {

			private static final GrpcVorgangHeader VORGANG = GrpcVorgangHeaderTestFactory.create();
			private static final GrpcFindVorgangResponse RESPONSE = GrpcFindVorgangResponse.newBuilder()
					.addVorgang(VORGANG).build();

			@BeforeEach
			void init() {
				when(stub.withInterceptors(any())).thenReturn(stub);
				when(stub.findVorgang(any())).thenReturn(RESPONSE);
			}

			@Test
			void shouldCallStub() {
				service.search(SEARCH_STR);

				verify(stub).findVorgang(notNull());
			}

			@Test
			void shouldMapResponse() {
				service.search(SEARCH_STR);

				verify(stubMapper).fromGrpc(notNull());
			}

			@Test
			void shouldReturnResult() {
				var vorgang = OzgCloudVorgangStubTestFactory.create();
				when(stubMapper.fromGrpc(any())).thenReturn(vorgang);

				var result = service.search(SEARCH_STR);

				assertThat(result).contains(vorgang);
			}
		}

	}

	@Nested
	class TestFind {

		private static final Page PAGE = Page.startPage();
		private final OzgCloudVorgangQuery query = OzgCloudVorgangQueryTestFactory.create();
		private final GrpcFindVorgangRequest request = GrpcFindVorgangRequestTestFactory.create();
		private final GrpcFindVorgangResponse response = GrpcFindVorgangResponseTestFactory.create();

		@Mock
		private VorgangServiceBlockingStub stubWithInterceptor;

		@BeforeEach
		void mock() {
			doReturn(stubWithInterceptor).when(service).getVorgangServiceStub();
			doReturn(request).when(service).buildFindRequest(query, PAGE);
			when(stubWithInterceptor.findVorgang(request)).thenReturn(response);
		}

		@Test
		void shouldCallGetVorgangServiceStub() {
			callFind();

			verify(service).getVorgangServiceStub();
		}

		@Test
		void shouldBuildFindRequest() {
			callFind();

			verify(service).buildFindRequest(query, PAGE);
		}

		@Test
		void shouldFindVorgang() {
			callFind();

			verify(stubWithInterceptor).findVorgang(request);
		}

		@Test
		void shouldMapResponse() {
			callFind();

			verify(stubMapper).fromGrpc(GrpcFindVorgangResponseTestFactory.VORGANG_HEADER);
		}

		@Test
		void shouldReturnMappedVorgangStubs() {
			var mappedVorgangStub = OzgCloudVorgangStubTestFactory.create();
			when(stubMapper.fromGrpc(GrpcFindVorgangResponseTestFactory.VORGANG_HEADER)).thenReturn(mappedVorgangStub);

			var vorgangStubs = callFind();

			assertThat(vorgangStubs).containsExactly(mappedVorgangStub);
		}

		private List<OzgCloudVorgangStub> callFind() {
			return service.find(query, PAGE);
		}
	}

	@Nested
	class TestBuildFindRequest {

		private final Page page = Page.builder()
				.limit(GrpcFindVorgangRequestTestFactory.LIMIT)
				.offset(GrpcFindVorgangRequestTestFactory.OFFSET)
				.build();

		private final OzgCloudVorgangQuery ozgCloudVorgangQuery = OzgCloudVorgangQueryTestFactory.create();

		@BeforeEach
		void mockMapper() {
			when(queryMapper.toGrpc(ozgCloudVorgangQuery)).thenReturn(GrpcFindVorgangRequestTestFactory.QUERY);
		}

		@Test
		void shouldMapToGrpcQuery() {
			callBuildFindRequest();

			verify(queryMapper).toGrpc(ozgCloudVorgangQuery);
		}

		@Test
		void shouldReturnRequest() {
			var request = callBuildFindRequest();

			assertThat(request).usingRecursiveComparison().isEqualTo(GrpcFindVorgangRequestTestFactory.create());
		}

		private GrpcFindVorgangRequest callBuildFindRequest() {
			return service.buildFindRequest(ozgCloudVorgangQuery, page);
		}
	}
}
